package provide Judy 1.0

critcl::clibraries -lJudy  -I. -I/usr/include/tcl${tcl_version}/tcl-private/generic

# this includes the debian copies of tclInt.h
critcl::cheaders /usr/include/tcl${tcl_version}/tcl-private/generic/tclInt.h /usr/include/tcl${tcl_version}/tcl-private/generic/tclIntDecls.h

# first include the basic Judy TclObj definition
critcl::ccode {
    #include "judyObj.c"
}

critcl::cinit {
    Tcl_RegisterObjType(&JudyType);
} {}

critcl::cproc judy_put {Tcl_Interp* interp Tcl_Obj* judyPtr Tcl_Obj* keyPtr Tcl_Obj* valuePtr} ok {
    Judy *judy;
    Tcl_Obj **value;
    char *key;
    PJError_t PJError;
    
    //fprintf(stderr, "judy_put: %04x\n", judyPtr);
    if (Tcl_IsShared(judyPtr)) {
	judyPtr = Tcl_DuplicateObj(judyPtr);
    }
    
    if (judyPtr->typePtr != &JudyType) {
	int result = SetJudyFromAny(interp, judyPtr);
	if (result != TCL_OK) {
	    //fprintf(stderr, "judy_put err: %04x\n", judyPtr);
	    return result;
	}
    }
    
    if (judyPtr->bytes != NULL) {
	Tcl_InvalidateStringRep(judyPtr);
    }

    judy = (Judy *) judyPtr->internalRep.otherValuePtr;

    key = Tcl_GetString(keyPtr);
    value = (Tcl_Obj **)JudySLGet(judy->table, key, PJError);
    if (value) {
	Tcl_DecrRefCount(*value);
    } else {
	//fprintf(stderr, "judy_put adding: %04x\n", judyPtr);
	judy->count++;
	value = (Tcl_Obj **)JudySLIns(&judy->table, key, PJError);
	//fprintf(stderr, "judy_put added: %04x\n", value);
    }
    *value = valuePtr;
    //fprintf(stderr, "judy_put incref: %04x\n", valuePtr);
    Tcl_IncrRefCount(valuePtr);
    //fprintf(stderr, "judy_put set: %04x\n", value);

    judy->maxkey = max(judy->maxkey, strlen(key));
    judy->epoch++;

    Tcl_SetObjResult(interp, judyPtr);
    //fprintf(stderr, "judy_put done: %04x\n", judyPtr);
    return TCL_OK;
}

critcl::cproc judy_remove {Tcl_Interp* interp Tcl_Obj* judyPtr Tcl_Obj* keyPtr} ok {
    Judy *judy;
    PJError_t PJError;
    char *key;
    Tcl_Obj **value;

    if (Tcl_IsShared(judyPtr)) {
	judyPtr = Tcl_DuplicateObj(judyPtr);
    }

    if (judyPtr->typePtr != &JudyType) {
	int result = SetJudyFromAny(interp, judyPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }

    if (judyPtr->bytes != NULL) {
	Tcl_InvalidateStringRep(judyPtr);
    }
    judy = (Judy *) judyPtr->internalRep.otherValuePtr;
    key = Tcl_GetString(keyPtr);
    value = (Tcl_Obj **)JudySLGet(judy->table, key, PJError);

    if (value) {
	Tcl_DecrRefCount(*value);
	JudySLDel(&judy->table, key, PJError);
	judy->epoch++;
	judy->count--;
    }

    Tcl_SetObjResult(interp, judyPtr);
    return TCL_OK;
}

critcl::cproc judy_get {Tcl_Interp* interp Tcl_Obj* judyPtr Tcl_Obj* keyPtr} ok {
    Judy *judy;
    Tcl_Obj **value;
    PJError_t PJError;

    if (judyPtr->typePtr != &JudyType) {
	int result = SetJudyFromAny(interp, judyPtr);
	fprintf(stderr, "judy_get convert: %04x\n", judyPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }

    judy = (Judy *) judyPtr->internalRep.otherValuePtr;
    value = (Tcl_Obj **)JudySLGet(judy->table, Tcl_GetString(keyPtr), PJError);

    if (value == NULL) {
	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("No such element", -1));
	return TCL_ERROR;
    } else {
	Tcl_SetObjResult(interp, judyPtr);
	return TCL_OK;
    }
}

# judy_create
critcl::ccommand judy_create {dummy ip objc objv} {
    Tcl_Obj *judyObj;
    int i;
    Judy *judy;

    /*
     * Must have an even number of arguments; note that number of
     * preceding arguments (i.e. "judy create" is also even, which
     * makes this much easier.)
     */
    if (!(objc & 1)) {
	Tcl_WrongNumArgs(ip, 2, objv, "?key value ...?");
	return TCL_ERROR;
    }
    
    judyObj = Tcl_NewJudyObj();
    //fprintf(stderr, "judyObj: %04x\n", judyObj);

    for (i=1 ; i<objc ; i+=2) {
	/*
	 * The next command is assumed to never fail...
	 */
	c_judy_put(ip, judyObj, objv[i], objv[i+1]);
    }
    Tcl_SetObjResult(ip, judyObj);
    return TCL_OK;
}
